from flor import flags
from flor.constants import *
from flor.logger import exp_json

from typing import List, Union, Optional

NO_INIT = None


class Window:
    def __init__(self, iterations_count: int, sparse_checkpoints: List[int]):
        self.iterations_count = exp_json.get("EPOCHS")
        self.extended_sparse_checkpoints: List[Optional[int]] = list(sparse_checkpoints)
        self.extended_sparse_checkpoints.insert(0, NO_INIT)

    def get_work_segment(self):
        if self._is_sparse():
            # Only a subset of epochs are valid entries
            # Only weak initialization is possible
            assert flags.MODE is REPLAY_MODE.weak
            lo = self._sparse()
            hi = self._sparse(hi=True)
            assert hi is not None
            return [
                Capsule(True, lo),
            ] + [Capsule(False, e) for e in range(lo + 1 if lo is not None else 0, hi)]
        else:
            # All epochs are valid entries
            our_segment = (
                self._get_segment_helper(self.iterations_count)[flags.PID.pid - 1]
                if flags.PID.pid > 0
                else []
            )
            if flags.MODE is REPLAY_MODE.weak:
                # Asks to initialize predecessor
                return [
                    Capsule(True, self._dense()),
                ] + [Capsule(False, e) for e in our_segment]
            else:
                idx = our_segment[0]
                assert idx is not None
                return [Capsule(True, e) for e in range(idx)] + [
                    Capsule(False, e) for e in our_segment
                ]

    def get_resume_epoch(self):
        """
        i : the epoch from which to resume (weak initialization)
        """
        if flags.MODE is REPLAY_MODE.strong:
            return None
        if self._is_sparse():
            return self._sparse()
        else:
            return self._dense()

    def _is_sparse(self):
        is_sparse = len(self.extended_sparse_checkpoints) > 1
        assert (
            is_sparse
            or len(self.extended_sparse_checkpoints) == 1
            and self.extended_sparse_checkpoints[0] is NO_INIT
        )
        return is_sparse

    def _get_segment_helper(self, num_resume_points):
        segments: List[List[Optional[int]]] = [[] for _ in range(flags.PID.ngpus)]
        for pt in range(num_resume_points):
            segments[pt % flags.PID.ngpus].append(None)
        i = 0
        for j in range(flags.PID.ngpus):
            for k in range(len(segments[j])):
                segments[j][k] = i
                i += 1
        assert i == num_resume_points
        return segments

    def _sparse(self, hi=False) -> Optional[int]:
        our_segment = self._get_segment_helper(len(self.extended_sparse_checkpoints))[
            flags.PID.pid - 1
        ]
        # TODO: ...
        assert (
            our_segment
        ), "TODO: Handle case when user allocs more partitions than there is work."
        # Re-index the segment with respect to self.sparse_checkpoints
        if not hi:
            assert our_segment[0] is not None
            return self.extended_sparse_checkpoints[our_segment[0]]
        else:
            if flags.PID.pid == flags.PID.ngpus:
                # This is the last segment
                return self.iterations_count
            else:
                # There exists a greater segment
                idx = our_segment[-1]
                assert idx is not None
                return self.extended_sparse_checkpoints[idx + 1]

    def _dense(self) -> Optional[int]:
        if flags.PID.pid == 0:
            assert isinstance(self.iterations_count, int) and self.iterations_count > 0
            return self.iterations_count - 1
        our_segment = self._get_segment_helper(self.iterations_count)[flags.PID.pid - 1]
        # TODO: ...
        assert (
            our_segment
        ), "TODO: Handle case when user allocs more partitions than there is work."
        pred_epoch = our_segment[0] - 1 if our_segment[0] else NO_INIT
        return pred_epoch


class Capsule:
    def __init__(self, init_only: bool, epoch: Optional[int]):
        self.init_only = init_only
        self.epoch = epoch



import matplotlib.pyplot as plt
from sklearn.metrics import confusion_matrix
import seaborn as sn
import pandas as pd
import numpy as np

from .shelf import home_shelf


def plot_confusion_matrix(y_true, y_pred, x_classes, y_classes):
    cf_matrix = confusion_matrix(y_true, y_pred)
    for i in range(len(cf_matrix)):
        cf_matrix[i][i] = 0
    df_cm = pd.DataFrame(cf_matrix, index=y_classes, columns=x_classes)
    plt.figure(figsize=(12, 7))
    sn.heatmap(df_cm, annot=False)

    path = home_shelf.get_ref(".png")
    plt.savefig(path)
    return path



import sqlite3
from pathlib import Path
from typing import Optional

from flor.shelf import home_shelf
from flor.state import State
from flor.constants import *

SUFFIX = ".db"
dbp: Optional[Path] = None


def exists():
    return dbp is not None and dbp.exists()


def start_db(projid: str):
    global dbp
    dir_n = "_".join(projid.split("_")[0:-1])
    dbp = home_shelf.florin / Path(dir_n).with_suffix(SUFFIX)
    is_first_start = not dbp.exists()
    assert State.db_conn is None, "Did you try to start_db more than once?"
    State.db_conn = sqlite3.connect(dbp)
    if is_first_start:
        init_db()


def init_db():
    """
    Initializes the database
    """
    assert State.db_conn is not None
    cur = State.db_conn.cursor()
    cur.executescript(
        """
        BEGIN;
        CREATE TABLE log_records(
            projid text,
            runid text,
            tstamp text,
            vid text, 
            epoch integer,
            step integer,
            name text,
            value text
            );
        CREATE TABLE data_prep(
            projid text,
            runid text,
            tstamp text,
            vid text,
            seconds real
        );
        CREATE TABLE outr_loop(
            projid text,
            runid text,
            tstamp text,
            vid text,
            epoch integer,
            seconds real
        );
        COMMIT;
        """
    )
    cur.close()


def get_watermark():
    assert State.db_conn is not None
    cur = State.db_conn.cursor()
    res = cur.execute("SELECT DISTINCT vid FROM log_records").fetchall()
    if not res:
        return None
    cur.close()
    return [str(row[0]) for row in res]


def get_schedule(keys):
    assert State.db_conn is not None
    cur = State.db_conn.cursor()
    if keys == DATA_PREP:
        res = []
        for r in cur.execute(
            "SELECT " + ", ".join(DATA_PREP) + ", seconds FROM data_prep;"
        ).fetchall():
            res.append(
                {
                    c: v
                    for c, v in zip(
                        list(DATA_PREP)
                        + [
                            "seconds",
                        ],
                        r,
                    )
                }
            )
    elif keys == OUTR_LOOP:
        res = []
        for r in cur.execute(
            "SELECT " + ", ".join(OUTR_LOOP) + ", seconds FROM outr_loop;"
        ).fetchall():
            res.append(
                {
                    c: v
                    for c, v in zip(
                        list(OUTR_LOOP)
                        + [
                            "seconds",
                        ],
                        r,
                    )
                }
            )
    else:
        raise
    if not res:
        return None
    cur.close()
    return res


def get_log_records():
    assert State.db_conn is not None
    cur = State.db_conn.cursor()
    res = cur.execute("SELECT * FROM log_records").fetchall()
    cur.close()
    return res


def write_log_records(list_of_dicts):
    assert State.db_conn is not None
    cur = State.db_conn.cursor()

    cur.executemany(
        "INSERT INTO log_records VALUES(?, ?, ?, ?, ?, ?, ?, ?)",
        [
            (
                d["projid"],
                d["runid"],
                d["tstamp"],
                d["vid"],
                d["epoch"],
                d["step"],
                d["name"],
                d["value"],
            )
            for d in list_of_dicts
        ],
    )
    State.db_conn.commit()
    print("Flor wrote log records to SqliteDB")
    cur.close()



from flor.shelf import home_shelf, cwd_shelf
from flor.constants import *
from flor.state import State
from flor.logger import exp_json
from flor.query.database import start_db

import sys
from typing import Dict, Optional, Tuple, Union
from pathlib import PurePath, Path
from time import time

NAME: Optional[str] = None
REPLAY: bool = False
INDEX: Optional[PurePath] = None
MODE: Optional[REPLAY_MODE] = None
PID: REPLAY_PARALLEL = REPLAY_PARALLEL(1, 1)
EPSILON: float = 1 / 15
RESUMING: bool = False

DATALOGGING = True

"""
--flor NAME [EPSILON]
--replay_flor [weak | strong] [i/n]
"""


def set_REPLAY(
    name: str,
    index: Optional[str] = None,
    mode: Optional[str] = None,
    pid: Optional[Tuple[int, int]] = None,
):
    """
    When set: enables FLOR REPLAY
    """
    global NAME, REPLAY, INDEX, MODE, PID
    NAME = name
    REPLAY = True
    MODE = REPLAY_MODE.weak
    if index is not None:
        assert isinstance(index, str)
        assert PurePath(index).suffix == ".json"
        assert home_shelf.verify(PurePath(index).name)
        INDEX = Path(index)
    if mode is not None:
        MODE = REPLAY_MODE[mode]
    if pid is not None:
        assert isinstance(pid, tuple) and len(pid) == 2
        p, n = pid
        assert isinstance(p, int) and isinstance(n, int)
        assert p >= 0
        assert n >= 1
        assert p <= n
        PID = REPLAY_PARALLEL(*pid)


class Parser:
    """
    --flor NAME [EPSILON]
    --replay_flor [weak | strong] [i/n]
    """

    @staticmethod
    def _parse_name():
        assert (
            "--replay_flor" not in sys.argv
        ), "Pick at most one of `--flor` or `--replay_flor` but not both"
        assert (
            cwd_shelf.in_shadow_branch()
        ), "Please invoke --flor from a `flor.shadow` branch."
        global NAME, EPSILON, DATALOGGING
        flor_flags = []
        feeding = False
        for _ in range(len(sys.argv)):
            arg = sys.argv.pop(0)

            if arg == "--flor":
                feeding = True
            elif arg[0] == "-":
                feeding = False

            if feeding:
                flor_flags.append(arg)
            else:
                sys.argv.append(arg)

        assert len(flor_flags) <= 3
        if flor_flags:
            assert flor_flags.pop(0) == "--flor"
            assert (
                flor_flags or exp_json.exists()
            ), "Missing NAME argument in --flor NAME"
            if exp_json.exists():
                exp_json.deferred_init()
            for flag in flor_flags:
                if flag[0:2] == "0.":
                    EPSILON = float(flag)
                else:
                    NAME = flag
            if NAME is None:
                assert exp_json.exists()
                NAME = exp_json.get("NAME")  # take from past
        assert NAME is not None
        if exp_json.exists() and exp_json.get("NAME") == NAME:
            # IF previous name is same as this name
            DATALOGGING = False
        exp_json.put("NAME", NAME)
        home_shelf.mk_job(cwd_shelf.get_projid())

    @staticmethod
    def _parse_replay():
        assert (
            "--flor" not in sys.argv
        ), "Pick at most one of `--flor` or `--replay_flor` but not both"
        assert (
            cwd_shelf.in_shadow_branch()
        ), "Please invoke --replay_flor from a `flor.shadow` branch."
        try:
            assert exp_json.exists()
            exp_json.deferred_init()
        except FileNotFoundError:
            print("No replay file, did you record first?")
            raise
        assert exp_json.get("NAME"), f"check your `{LOG_RECORDS}` file. Missing name."
        flor_flags = []
        feeding = False
        for _ in range(len(sys.argv)):
            arg = sys.argv.pop(0)

            if arg == "--replay_flor":
                feeding = True
            elif arg[0] == "-":
                feeding = False

            if feeding:
                flor_flags.append(arg)
            else:
                sys.argv.append(arg)

        assert len(flor_flags) <= 3
        if flor_flags:
            assert flor_flags.pop(0) == "--replay_flor"
            mode = None
            pid = None
            for flor_flag in flor_flags:
                if flor_flag in [each.name for each in REPLAY_MODE]:
                    mode = flor_flag
                elif len(flor_flag.split("/")) == 2:
                    p, n = flor_flag.split("/")
                    pid = int(p), int(n)
                else:
                    raise RuntimeError(
                        "Invalid argument passed to --replay_flor"
                        + "[weak | strong] [I/N]"
                    )
            projid = cwd_shelf.get_projid()
            set_REPLAY(
                str(exp_json.get("NAME")),
                index=exp_json.get("MEMO"),
                mode=mode,
                pid=pid,
            )
            home_shelf.set_job(projid)
            start_db(projid)

    @staticmethod
    def parse():
        State.import_time = time()
        if "--flor" in sys.argv:
            Parser._parse_name()
        elif "--replay_flor" in sys.argv:
            Parser._parse_replay()


__all__ = [
    "NAME",
    "REPLAY",
    "INDEX",
    "MODE",
    "PID",
    "EPSILON",
    "DATALOGGING",
    "set_REPLAY",
    "Parser",
]



from .constants import *
from abc import ABC, abstractmethod

import json


class Entry(ABC):
    next_lsn = 0

    def __init__(self, sk, gk):
        self.sk = sk
        self.gk = gk
        self.lsn = Entry.next_lsn
        Entry.next_lsn += 1

    def jsonify(self):
        d = dict()
        d[STATIC_KEY] = str(self.sk)
        d[GLOBAL_KEY] = int(self.gk)
        d[GLOBAL_LSN] = int(self.lsn)
        return d

    @abstractmethod
    def is_left(self):
        ...

    @abstractmethod
    def is_right(self):
        ...

    @staticmethod
    @abstractmethod
    def is_superclass(json_dict: dict):
        ...

    @classmethod
    @abstractmethod
    def cons(cls, json_dict: dict):
        ...



from .value import Value as DataVal
from .reference import Reference as DataRef
from .dataframe import DataFrame

from typing import Type, Union

d_entry = Union[DataVal, DataRef, DataFrame]

__all__ = ["DataVal", "DataRef", "DataFrame", "d_entry"]

try:
    import torch
    from .torch_chkpt import Torch

    t_d_entry = Union[d_entry, Torch]
    __all__.extend(["Torch", "t_d_entry"])

except ModuleNotFoundError:
    pass



from dataclasses import dataclass
from enum import Enum
from pathlib import Path

REPLAY_MODE = Enum("REPLAY_MODE", "weak strong")
SHADOW_BRANCH_PREFIX = "flor.shadow"
LOG_RECORDS = Path(".flor") / "log_records.csv"
REPLAY_JSON = Path(".flor") / ".replay.json"
SECONDS_JSON = Path(".flor") / "seconds.json"

DATA_PREP = (
    "projid",
    "runid",
    "tstamp",
    "vid",
)
OUTR_LOOP = DATA_PREP + ("epoch",)
INNR_LOOP = OUTR_LOOP + ("step",)

DESERIALIZATION_COEFF = 3.573e-08


@dataclass
class REPLAY_PARALLEL:
    pid: int
    ngpus: int



import uuid
from datetime import datetime
from pathlib import Path, PurePath
from typing import Optional, Union
import json
import os

from flor import flags
from flor.state import State
from flor.logger import exp_json
from flor.skipblock import SkipBlock

home: Path = Path.home()

florin: Path = home / ".flor"
florin.mkdir(exist_ok=True)

job: Optional[Path] = None
data: Optional[Path] = None
timestamp: Optional[str] = None


def get_checkpoint_bytes_per_epoch(projid: str):
    pkl_paths = []
    with open(florin / projid / "latest.json", "r") as f:
        for jstring in f.readlines():
            d = json.loads(jstring)
            if "metadata" in d and d["metadata"] == "LBRACKET" and pkl_paths:
                break
            elif "metadata" in d and d["metadata"] == "LBRACKET":
                continue
            elif "torch_ref" in d:
                pkl_paths.append(d["torch_ref"])
    return sum([os.path.getsize(florin / projid / "data" / p) for p in pkl_paths])


def mk_job(name: str):
    global timestamp, job, data
    assert isinstance(name, str)
    timestamp = datetime.now().isoformat()
    exp_json.put("TSTAMP", str(PurePath(timestamp).with_suffix(".json")))
    State.timestamp = timestamp
    job = florin / name
    job.mkdir(exist_ok=True)
    data = job / "data"
    data.mkdir(exist_ok=True)
    (job / "csv").mkdir(exist_ok=True)


def set_job(name: str):
    global timestamp, job, data
    assert isinstance(name, str)
    timestamp = exp_json.get("TSTAMP")
    State.timestamp = timestamp
    job = florin / name
    data = job / "data"


def get_index() -> Optional[Path]:
    if job is not None and timestamp is not None:
        if flags.REPLAY:
            projid = exp_json.get("PROJID")
            tstamp = exp_json.get("TSTAMP")
            assert projid is not None and tstamp is not None
            return florin / PurePath(projid) / PurePath(tstamp)
        else:
            return job / PurePath(timestamp).with_suffix(".json")
    else:
        return None


def get_job():
    assert job is not None
    return job


def get_latest() -> Optional[Path]:
    return job / PurePath("latest").with_suffix(".json") if job is not None else None


def get_ref(ext=".pkl"):
    assert data is not None
    return data / PurePath(uuid.uuid4().hex).with_suffix(ext)


def get_pkl_ref() -> Optional[Path]:
    return get_ref(".pkl")


def get_csv_ref() -> Optional[Path]:
    return get_ref(".csv")


def verify(path: Union[PurePath, str]) -> bool:
    assert flags.NAME is not None
    resolved_path = florin / flags.NAME / path
    return resolved_path.exists()


def close():
    if not flags.REPLAY and flags.NAME:
        if len(SkipBlock.logger.buffer) > 0 or SkipBlock.logger.flush_count:
            path = get_index()
            assert path is not None
            if len(SkipBlock.logger.buffer) > 0:
                SkipBlock.logger.flush(is_final=True)
            latest = get_latest()
            assert latest is not None
            assert path.exists()
            if latest.exists():
                latest.unlink()
            latest.symlink_to(path)
            return path



from .block import Block
from .group import BlockGroup
from .window import Window
from ..entry import *

from collections import OrderedDict as ODict
from typing import List, Optional, Union, Dict


class Tree:
    def __init__(self, log_entry: Optional[Bracket] = None):
        self.hash: Dict[str, BlockGroup] = ODict()
        self.block = None

        if log_entry is not None:
            assert log_entry.is_left()
            self.block = Block(log_entry)
            self._hash(self.block)

        self.root = self.block

        self.sparse_checkpoints: List[int] = []
        self.iterations_count = 0

    def __getitem__(self, item) -> BlockGroup:
        return self.hash[item]

    def _hash(self, block: Block):
        if block.static_key in self.hash:
            self.hash[block.static_key].add_block(block)
        else:
            self.hash[block.static_key] = BlockGroup(block)

    def add_sparse_checkpoint(self):
        # print(f"SPARSE CHECKPOINT at {self.iterations_count - 1}")
        self.sparse_checkpoints.append(self.iterations_count - 1)

    def feed_entry(self, log_entry: Union[d_entry, md_entry]):
        if self.root is None:
            assert self.block is None
            assert log_entry.is_left()
            assert isinstance(log_entry, Bracket)
            self.block = Block(log_entry)
            self.root = self.block
            self._hash(self.block)
        elif log_entry.is_left():
            assert self.block is not None
            if self.block.right_fed:
                assert isinstance(log_entry, Bracket)
                successor = Block(log_entry, self.block.parent)
                self.block.successor = successor
                self.block = successor
                self._hash(successor)
            else:
                assert isinstance(log_entry, Bracket)
                child = Block(log_entry, self.block)
                self.block.child = child
                self.block = child
                self._hash(child)
        elif log_entry.is_right():
            assert self.block is not None
            assert not isinstance(log_entry, EOF)
            if self.block.belongs_in_block(log_entry):
                self.block.feed_entry(log_entry)
            else:
                assert (
                    self.block.parent is not None
                    and self.block.parent.belongs_in_block(log_entry)
                )
                self.block = self.block.parent
                self.block.feed_entry(log_entry)
        else:
            assert isinstance(log_entry, EOF)
            self.sparse_checkpoints = log_entry.sparse_checkpoints
            self.iterations_count = log_entry.iterations_count

        if log_entry.is_left() and log_entry.sk == self.root.static_key:
            self.iterations_count += 1

    def parse(self, records: List[Union[d_entry, md_entry]]):
        assert self.block is None, "Need a new Tree for parsing"
        for rec in records:
            self.feed_entry(rec)

    def get_resume_epoch(self):
        """
        Gets the predecessor epoch for weak initialization,
            if such epoch exists, else None.
        """
        return Window(self.iterations_count, self.sparse_checkpoints).get_resume_epoch()

    def get_segment(self):
        """
        Gets the relevant work segment, including tagged initialization epochs
        """
        return Window(self.iterations_count, self.sparse_checkpoints).get_work_segment()



from ..constants import *
from .abstract import Data

import json


class Value(Data):
    def __init__(self, sk, gk, v):
        super().__init__(sk, gk, v)

    def would_mat(self):
        """
        For timing serialization costs
        """
        d = self.jsonify()
        json.dumps(d)

    def make_val(self):
        pass

    @staticmethod
    def is_superclass(json_dict: dict):
        return VAL in json_dict

    @classmethod
    def cons(cls, json_dict: dict):
        return cls(json_dict[STATIC_KEY], json_dict[GLOBAL_KEY], json_dict[VAL])

    def promise(self):
        self.promised = self.value

    def fulfill(self):
        super().fulfill()
        return json.dumps(self.jsonify())



from typing import List, Dict, Set
from enum import Enum
from pathlib import Path
import pandas as pd
import ast
from time import time

from flor.constants import *
from flor.state import State
from flor.hlast import apply

from flor.hlast.visitors import LoggedExpVisitor

import subprocess


def get_dims(pivot_vars: Dict[str, Set[str]], apply_vars: List[str]):
    if any([applied_v in pivot_vars["INNR_LOOP"] for applied_v in apply_vars]):
        return INNR_LOOP
    if any([applied_v in pivot_vars["OUTR_LOOP"] for applied_v in apply_vars]):
        return OUTR_LOOP
    return DATA_PREP


def batch_replay(apply_vars: List[str], path: str, versions: pd.Series, loglvl):
    # TODO: argv processing
    base_cmd = ["python", path, "--replay_flor"]

    start_time = time()
    assert State.repo is not None
    for hexsha in versions:
        State.repo.git.checkout(hexsha)
        apply(diff_vars(apply_vars, path), path)

        if loglvl == DATA_PREP:
            subprocess.run(
                base_cmd
                + [
                    "0/1",
                ]
            )
        elif loglvl == OUTR_LOOP:
            subprocess.run(base_cmd)
        elif loglvl == INNR_LOOP:
            subprocess.run(
                base_cmd
                + [
                    "1/2",
                ]
            )
            subprocess.run(
                base_cmd
                + [
                    "2/2",
                ]
            )
        else:
            raise
        State.repo.git.stash()
    print(f"Time elapsed: {time() - start_time} seconds")


def diff_vars(apply_vars: List[str], path: str):
    with open(path, "r") as f:
        tree = ast.parse(f.read())
    visitor = LoggedExpVisitor()
    visitor.visit(tree)
    return [v for v in apply_vars if v not in visitor.names]


__all__ = ["get_dims", "batch_replay"]



STATIC_KEY = "static_key"
GLOBAL_KEY = "global_key"
GLOBAL_LSN = "global_lsn"
VAL = "value"
REF = "ref"
LBRACKET = "LBRACKET"
RBRACKET = "RBRACKET"
EOF_NAME = "EOF"
METADATA = "metadata"
SPARSE_CHECKPOINTS = "sparse_checkpoints"
ITERATIONS_COUNT = "iterations_count"
PKL_SFX = ".pkl"



from typing import Iterable, List, Union

from flor import flags
from flor.shelf import home_shelf as shelf, cwd_shelf
from flor.skipblock import SkipBlock

from flor.constants import *

from flor.state import State


runtime_initialized = False


def it(value: Union[Iterable, bool]):
    """
    Main loop wrapper
    :param value:
        Iterable when iterating over a for loop
        Bool when looping with while
    """
    assert isinstance(value, (Iterable, bool))
    if flags.NAME is None:
        if isinstance(value, bool):
            return value
        else:
            assert isinstance(value, Iterable)
            for each in value:
                yield each
            return

    _deferred_init()

    if not flags.REPLAY:
        # Record mode
        if isinstance(value, bool):
            return value
        else:
            for each in value:
                yield each
    else:
        # Replay mode
        segment = SkipBlock.journal.get_segment_window()
        for capsule in segment:
            flags.RESUMING = capsule.init_only
            if isinstance(value, bool):
                yield True
            else:
                assert isinstance(value, Iterable)
                if flags.RESUMING:
                    if capsule.epoch is None:
                        continue
                    else:
                        State.epoch = int(capsule.epoch)
                        assert hasattr(
                            value, "__getitem__"
                        ), "TODO: Implement next() calls to consume iterator"
                        yield value[capsule.epoch]  # type: ignore
                else:
                    assert capsule.epoch is not None
                    State.epoch = value[capsule.epoch]  # type: ignore
                    assert hasattr(
                        value, "__getitem__"
                    ), "TODO: Implement next() calls to consume iterator"
                    yield value[capsule.epoch]  # type: ignore


def _deferred_init(_nil=[]):
    """
    At most once execution
    """
    global runtime_initialized
    if not runtime_initialized:
        if flags.NAME is not None:
            assert (
                cwd_shelf.in_shadow_branch()
            ), f"Please run FLOR from a shadow branch (branch name: `{SHADOW_BRANCH_PREFIX}.[...]`)\n"
            SkipBlock.bind()
            if flags.REPLAY:
                SkipBlock.journal.read()
        if flags.NAME and not flags.REPLAY:
            index_path = shelf.get_index()
            SkipBlock.logger.set_path(index_path)

        runtime_initialized = True


__all__ = ["it"]



from git.repo import Repo
from git.exc import InvalidGitRepositoryError
import os

from flor import flags
from flor.state import State

from flor.constants import *
from flor.logger import exp_json, log_records
from flor.shelf import home_shelf
from pathlib import Path

import atexit

PATH = Path(".flor")


def get_projid():
    if State.common_dir is None:
        r = Repo()
        State.common_dir = Path(r.common_dir)
    return (
        os.path.basename(os.path.dirname(str(State.common_dir)))
        + "_"
        + str(State.active_branch)
    )


def in_shadow_branch():
    """
    Initialize
    """
    try:
        r = Repo()
        State.repo = r
        try:
            State.active_branch = str(r.active_branch)
        except TypeError:
            bs = State.repo.git.branch(
                "--contains", str(State.repo.head.commit.hexsha)
            ).split("\n")
            branches = [e.strip() for e in bs if "flor.shadow" in e]
            if len(branches) == 1:
                State.active_branch = branches.pop()
            else:
                raise NotImplementedError()
        cond = check_branch_cond()
        if cond:
            PATH.mkdir(exist_ok=True)
            get_projid()
        return cond
    except InvalidGitRepositoryError:
        return False


def check_branch_cond():
    if State.repo is not None:
        return "RECORD::" in str(
            State.repo.head.commit.message
        ) or "flor.shadow" in State.repo.git.branch(
            "--contains", str(State.repo.head.commit.hexsha)
        )
    return False


@atexit.register
def flush():
    path = home_shelf.close()
    cond = in_shadow_branch()
    projid = get_projid()

    if flags.NAME and not flags.REPLAY:
        assert cond
        log_records.flush(projid, str(State.timestamp))
        exp_json.put("PROJID", projid)
        exp_json.put("EPOCHS", State.epoch)
        exp_json.flush()
        repo = State.repo
        assert repo is not None
        repo.git.add("-A")
        repo.index.commit(f"RECORD::{flags.NAME}")
    elif flags.NAME and flags.REPLAY:
        assert cond
        for k in [k for k in exp_json.record_d if not k.isupper()]:
            log_records.put_dp(k, exp_json.record_d[k])

        log_records.flush(projid, str(State.timestamp))

    if State.db_conn:
        State.db_conn.commit()
        State.db_conn.close()



import json
from typing import List, Union

from flor import flags
from flor.state import State
from flor.shelf import home_shelf
from flor.constants import *
from flor.logger import exp_json

from .entry import *
from .tree import Tree
from .tree.window import Capsule


def read_entries() -> List[Union[d_entry, md_entry]]:
    assert flags.REPLAY
    index = home_shelf.get_index()
    assert index is not None, "FileNotFound. Missing checkpoint directory, or job name"

    with open(index, "r") as f:
        return [make_entry(json.loads(line.strip())) for line in f]


class Journal:
    def __init__(self):
        self.tree = Tree()  # type: ignore
        self.sub_tree = None
        self.entries = None

    def read(self):
        self.entries = read_entries()
        self.tree.parse(self.entries)

    def get_segment_window(self) -> List[Capsule]:
        assert flags.PID.ngpus <= self.tree.iterations_count
        if self.tree.sparse_checkpoints:
            assert (
                flags.PID.ngpus <= len(self.tree.sparse_checkpoints) + 1
            ), f"Not enough checkpoints. Max degree of parallelism: {len(self.tree.sparse_checkpoints) + 1}"
        if flags.MODE is REPLAY_MODE.weak and flags.PID.pid >= 0:
            self._advance_head()
            assert self.sub_tree is not None
            return self.sub_tree.get_segment()
        return self.tree.get_segment()

    def get_iterations_count(self):
        return exp_json.get("EPOCHS")

    def as_tree(self) -> Tree:
        if (
            not flags.REPLAY
            or flags.MODE is REPLAY_MODE.strong
            or (flags.PID.pid == 1 and flags.PID.ngpus == 1)
        ):
            return self.tree
        else:
            assert self.sub_tree is not None
            return self.sub_tree

    def get_eof(self, commit_sha: str):
        tree = self.as_tree()
        return EOF(tree.sparse_checkpoints, tree.iterations_count, commit_sha)

    def _advance_head(self):
        """
        Used for checkpoint resume,
        ignores journal entries that precede the first epoch of work
        """
        self.sub_tree = Tree()  # type: ignore
        epoch_to_init: Union[int, None] = self.tree.get_resume_epoch()
        if epoch_to_init is not None:
            assert self.tree.root is not None
            target = self.tree[self.tree.root.static_key].blocks[epoch_to_init]
            State.target_block = target  # type: ignore
            feeding = False
            assert self.entries is not None
            for journal_entry in self.entries:
                if (
                    not feeding
                    and journal_entry.is_left()
                    and journal_entry.sk == target.static_key
                    and journal_entry.gk == target.global_key
                ):
                    feeding = True
                if feeding:
                    self.sub_tree.feed_entry(journal_entry)


__all__ = ["Journal"]



import csv
from typing import List, Set, Dict
import pandas as pd
import numpy as np
from flor.shelf import home_shelf, cwd_shelf
from flor.query.unpack import unpack, clear_stash
from flor.query import database
from flor.shelf import cwd_shelf
from pathlib import Path
from IPython.display import display

from flor.query.pivot import *
from flor.query.engine import *
from flor.constants import *
from flor.state import State

pivot_vars: Dict[str, Set[str]] = {
    "DATA_PREP": set([]),
    "OUTR_LOOP": set([]),
    "INNR_LOOP": set([]),
}


def log_records(skip_unpack=False):
    if not skip_unpack:
        unpack()
    else:
        # do unpack initialize
        assert cwd_shelf.in_shadow_branch()
        if State.db_conn is None:
            database.start_db(cwd_shelf.get_projid())

    facts = (
        pd.DataFrame(
            database.get_log_records(),
            columns=[
                "projid",
                "runid",
                "tstamp",
                "vid",
                "epoch",
                "step",
                "name",
                "value",
            ],
        )
        .astype(
            {
                "projid": str,
                "runid": str,
                "tstamp": np.datetime64,
                "vid": str,
                "epoch": int,
                "step": int,
                "name": str,
                "value": object,
            }
        )
        .sort_values(by=["tstamp", "epoch", "step"])
    )
    return facts


def full_pivot(facts: pd.DataFrame):
    data_prep_gb = facts.drop_duplicates()[list(DATA_PREP) + ["name", "value"]].groupby(
        by=list(DATA_PREP + ("name",))
    )
    for rowid, agg in data_prep_gb.count()["value"].items():
        name = str(tuple(rowid)[-1])  # type: ignore
        if agg == 1:
            pivot_vars["DATA_PREP"] |= {
                name,
            }

    outer_loop_gb = (
        facts[list(OUTR_LOOP) + ["name", "value"]]
        .drop_duplicates()
        .groupby(by=list(OUTR_LOOP + ("name",)))
    )
    for rowid, agg in outer_loop_gb.count()["value"].items():
        name = str(tuple(rowid)[-1])  # type: ignore
        if name not in pivot_vars["DATA_PREP"] and agg == 1:
            pivot_vars["OUTR_LOOP"] |= {
                name,
            }

    pivot_vars["INNR_LOOP"] |= set(
        [
            name
            for name in facts["name"]
            if name not in pivot_vars["DATA_PREP"]
            and name not in pivot_vars["OUTR_LOOP"]
        ]
    )

    dp_keys = DATA_PREP
    dp_pivot = data_prep_pivot(facts, pivot_vars["DATA_PREP"])

    ol_keys = OUTR_LOOP
    ol_pivot = outer_loop_pivot(facts, pivot_vars["OUTR_LOOP"])

    all_keys = INNR_LOOP
    il_pivot = inner_loop_pivot(facts, pivot_vars["INNR_LOOP"])

    def post_proc(df, df_keys):
        df_keys = list(df_keys)
        df_keys.extend([c for c in df.columns if c not in df_keys])
        return df[df_keys].sort_values(
            [k for k in df_keys if k in ("tstamp", "epoch", "step")]
        )

    if ol_pivot is not None and il_pivot is not None:
        ol_il = il_pivot.merge(ol_pivot, how="outer", on=ol_keys)
        if dp_pivot is not None:
            return post_proc(ol_il.merge(dp_pivot, how="outer", on=dp_keys), all_keys)
        return post_proc(ol_il, all_keys)
    elif dp_pivot is not None and ol_pivot is not None:
        return post_proc(ol_pivot.merge(dp_pivot, how="outer", on=dp_keys), ol_keys)
    elif dp_pivot is not None and il_pivot is not None:
        return post_proc(il_pivot.merge(dp_pivot, how="outer", on=dp_keys), all_keys)
    elif dp_pivot is not None:
        return post_proc(dp_pivot, dp_keys)
    elif ol_pivot is not None:
        return post_proc(ol_pivot, ol_keys)
    elif il_pivot is not None:
        return post_proc(il_pivot, all_keys)


def replay(apply_vars: List[str], where_clause: str, path: str):
    """
    apply_vars : ['device', 'optimizer', 'learning_rate', ...]
    where_clause: stated in Pandas/SQL, passed to full_pivot
    path: `train_rnn.py` or such denoting main python script
    """
    assert Path(path).suffix == ".py"
    facts = log_records(skip_unpack=True)
    df = full_pivot(facts)
    assert df is not None
    assert all([v in facts['name'].values for v in apply_vars])

    loglvl = get_dims(pivot_vars, apply_vars)
    dp_schedule = (
        df.query(where_clause)[list(DATA_PREP)]
        .drop_duplicates()
        .merge(
            pd.DataFrame(database.get_schedule(DATA_PREP)).astype(
                {
                    "projid": str,
                    "runid": str,
                    "tstamp": np.datetime64,
                    "vid": str,
                    "seconds": float,
                }
            ),
            how="inner",
            on=DATA_PREP,
        )[
            list(DATA_PREP)
            + [
                "seconds",
            ]
        ]
    )
    if loglvl == DATA_PREP:
        versions = dp_schedule["vid"].drop_duplicates()
        display(dp_schedule)
        print(
            f"Continue replaying {len(versions)} versions at DATA_PREP level for {'{:.2f}'.format(3 * len(versions) + sum(dp_schedule['seconds']))} seconds?"
        )
        res = input("Continue [Y/n]? ")
        if res.strip().lower() != "n":
            batch_replay(apply_vars, path, versions, DATA_PREP)
    elif loglvl == OUTR_LOOP:
        size_bytes = home_shelf.get_checkpoint_bytes_per_epoch(cwd_shelf.get_projid())
        schedule = (
            df.query(where_clause)[list(OUTR_LOOP)]
            .drop_duplicates()
            .merge(
                pd.DataFrame(database.get_schedule(OUTR_LOOP)).astype(
                    {
                        "projid": str,
                        "runid": str,
                        "tstamp": np.datetime64,
                        "vid": str,
                        "epoch": int,
                        "seconds": float,
                    }
                ),
                how="inner",
                on=OUTR_LOOP,
            )[
                list(OUTR_LOOP)
                + [
                    "seconds",
                ]
            ]
        )
        schedule["seconds"] = size_bytes * DESERIALIZATION_COEFF
        versions = schedule["vid"].drop_duplicates()
        display(schedule)
        print(
            f"Continue replaying {len(versions)} versions at OUTR_LOOP level for {'{:.2f}'.format((3 * len(versions) + sum(dp_schedule['seconds'])) + sum(schedule['seconds']))} seconds?"
        )
        res = input("Continue [Y/n]? ")
        if res.strip().lower() != "n":
            batch_replay(apply_vars, path, versions, OUTR_LOOP)
    elif loglvl == INNR_LOOP:
        schedule = (
            df.query(where_clause)[list(OUTR_LOOP)]
            .drop_duplicates()
            .merge(
                pd.DataFrame(database.get_schedule(OUTR_LOOP)).astype(
                    {
                        "projid": str,
                        "runid": str,
                        "tstamp": np.datetime64,
                        "vid": str,
                        "epoch": int,
                        "seconds": float,
                    }
                ),
                how="inner",
                on=OUTR_LOOP,
            )[
                list(OUTR_LOOP)
                + [
                    "seconds",
                ]
            ]
        )
        versions = schedule["vid"].drop_duplicates()
        display(schedule)
        print(
            f"Continue replaying {len(versions)} versions at INNR_LOOP level for {'{:.2f}'.format(sum(3 * len(versions) + dp_schedule['seconds']) + sum(schedule['seconds']))} seconds?"
        )
        res = input("Continue [Y/n]? ")
        if res.strip().lower() != "n":
            batch_replay(apply_vars, path, versions, INNR_LOOP)
    else:
        raise


__all__ = ["log_records", "full_pivot", "clear_stash", "replay"]



from flor.skipblock import SkipBlock
from flor.constants import *

from flor.state import State
from flor.iterator import it

from time import time
from typing import Iterable, Any


class MTK:
    """
    MODEL TRAINING KIT
    """

    SkipBlock = SkipBlock
    chckpts = []  # type: ignore

    @staticmethod
    def checkpoints(*args):
        MTK.chckpts.extend([a for a in args])

    @staticmethod
    def loop(iter8r, name=None, probed=None) -> Iterable[Any]:
        """
        Commits after every outer loop
        """
        try:
            State.loop_nesting_level += 1
            assert State.loop_nesting_level >= 1
            static_id = f"""{'outer loop' if State.loop_nesting_level == 1 else 'nested loop'}"""
            name = str(static_id) if name is None else name
            if State.loop_nesting_level == 1:
                # Outer loop
                assert State.import_time is not None
                State.seconds["PREP"] = time() - State.import_time
                State.seconds["EPOCHS"] = []
                State.epoch = 0
                for each in it(iter8r):
                    start_time = time()
                    State.step = 0
                    State.epoch += 1
                    yield each
                    State.seconds["EPOCHS"].append(time() - start_time)
            else:
                assert State.loop_nesting_level > 1
                # Nested loop
                if SkipBlock.step_into(name, probed):
                    assert State.step is not None
                    try:
                        while True:
                            each = next(iter8r)
                            State.step += 1
                            yield each
                    except StopIteration:
                        pass
                    finally:
                        SkipBlock.end(*MTK.chckpts)
                else:
                    SkipBlock.end(*MTK.chckpts)

        finally:
            State.loop_nesting_level -= 1



from typing import Dict, Any
from flor.logger.checkpoint_logger import Logger
from flor.logger.future import Future
from copy import deepcopy
import json


from flor.logger import exp_json, log_records
from flor.state import State
from flor import flags


def log(name, value):
    def is_jsonable(x):
        try:
            json.dumps(x)
            return True
        except (TypeError, OverflowError):
            return False

    serializable_value = value if is_jsonable(value) else str(value)
    if State.loop_nesting_level:
        log_records.put(name, serializable_value)
    else:
        exp_json.put(name, serializable_value, ow=False)
    return value


def pinned(name, callback, *args):
    if not flags.REPLAY:
        value = callback(*args)
        if flags.NAME:
            exp_json.put(name, value, ow=False)
    else:
        value = exp_json.get(name)
    return value



from typing import Dict, List, Optional, Union
from pathlib import Path
from git.repo import Repo
import sqlite3


class State:
    loop_nesting_level = 0
    epoch: Optional[int] = None
    step: Optional[int] = None
    timestamp: Optional[str] = None
    repo: Optional[Repo] = None
    common_dir: Optional[Path] = None
    active_branch: Optional[str] = None
    db_conn: Optional[sqlite3.Connection] = None
    import_time: Optional[float] = None
    seconds: Dict[str, Optional[Union[float, List[float]]]] = {
        "PREP": None,
    }  # May add "EPOCHS" at runtime
    target_block = None



from .seemblock import SeemBlock

from flor import flags
from flor import shelf
from flor.journal.entry import *

import pandas as pd

import time
from typing import Dict, List, Union


class WriteBlock(SeemBlock):
    scaling_factor = 1.38
    dynamic_identifiers: Dict[str, int] = dict()
    pda: List[Bracket] = []

    @staticmethod
    def step_into(block_name: str, probed=None):
        assert isinstance(block_name, str)
        dynamic_id = WriteBlock.dynamic_identifiers.get(block_name, 0)
        WriteBlock.dynamic_identifiers[block_name] = dynamic_id + 1

        lbracket = Bracket(
            block_name, dynamic_id, LBRACKET, predicate=True, timestamp=time.time()
        )
        WriteBlock.journal.as_tree().feed_entry(lbracket)

        WriteBlock.logger.append(lbracket)
        WriteBlock.pda.append(lbracket)
        return lbracket.predicate

    @staticmethod
    def end(*args, values=None):
        lbracket = WriteBlock.pda.pop()
        assert lbracket.timestamp is not None
        block_group = WriteBlock.journal.as_tree()[lbracket.sk]
        block = block_group.peek_block()
        assert block.global_key == lbracket.gk
        block_group.tick_execution(time.time() - lbracket.timestamp)
        if not args:
            rbracket = Bracket(lbracket.sk, lbracket.gk, RBRACKET)
            WriteBlock.journal.as_tree().feed_entry(rbracket)
            WriteBlock.logger.append(rbracket)
            block_group.set_mat_time(0)
        else:
            data_records = []

            start_time = time.time()
            for arg in args:
                data_record = val_to_record(arg, lbracket)
                data_records.append(data_record)
            block_group.set_mat_time(time.time() - start_time)

            if WriteBlock._should_materialize(block_group):
                for data_record in data_records:
                    WriteBlock.journal.as_tree().feed_entry(data_record)
                    WriteBlock.logger.append(data_record)
                block_group.tick_materialization()
            else:
                rbracket = Bracket(lbracket.sk, lbracket.gk, RBRACKET)
                WriteBlock.journal.as_tree().feed_entry(rbracket)
                WriteBlock.logger.append(rbracket)

    @staticmethod
    def _should_materialize(block_group):
        assert block_group.materialization_time is not None
        assert block_group.computation_time is not None

        block = block_group.peek_block()

        # Must align successor checkpoints for periodic checkpointing
        if block.force_mat:
            return True

        # First consider atomic case (always/never)
        ratio = block_group.materialization_time / block_group.computation_time
        threshold = min(1 / (1 + WriteBlock.scaling_factor), flags.EPSILON)  # type: ignore
        if ratio < threshold:
            return True

        # Then account for parallelism speedup
        if block.parent is None:
            threshold *= block_group.executions_count / (
                block_group.materializations_count + 1
            )
            if ratio < threshold:
                WriteBlock.journal.as_tree().add_sparse_checkpoint()
                block.force_mat_successors()
                return True

        return False


def val_to_record(arg, lbracket: Bracket) -> d_entry:
    if type(arg) in [type(None), int, float, bool, str]:
        return DataVal(lbracket.sk, lbracket.gk, arg)
    elif isinstance(arg, pd.DataFrame):
        return DataFrame(lbracket.sk, lbracket.gk, arg)
    else:
        if hasattr(arg, "state_dict"):
            try:
                return Torch(lbracket.sk, lbracket.gk, arg.state_dict())  # type: ignore
            except Exception as e:
                print("for debugging", e)
        return DataRef(lbracket.sk, lbracket.gk, arg)



from flor.constants import REPLAY_JSON, SECONDS_JSON
from flor import flags
from flor.state import State
from typing import Optional, Dict, Any

import json, os
from pathlib import Path

replay_d: Optional[Dict[str, Any]] = None
record_d = {}


def _get_path(p=REPLAY_JSON):
    assert State.common_dir is not None
    return os.path.join(os.path.dirname(State.common_dir), p)


def deferred_init():
    global replay_d
    if State.common_dir is not None:
        with open(_get_path(), "r", encoding="utf-8") as f:
            replay_d = json.load(f)


def exists():
    if State.common_dir is not None:
        path = Path(_get_path())
        return path.exists()
    else:
        return False


def put(name, value, ow=True):
    assert ow or name not in record_d
    record_d[name] = value


def get(name):
    """Get from previous run"""
    assert replay_d is not None
    return replay_d.get(name, None)


def flush():
    if flags.NAME and not flags.REPLAY:
        assert State.common_dir is not None
        with open(_get_path(), "w", encoding="utf-8") as f:
            json.dump(record_d, f, ensure_ascii=False, indent=4)
        
        assert State.seconds["PREP"] is not None
        with open(_get_path(SECONDS_JSON), "w", encoding="utf-8") as f:
            json.dump(State.seconds, f, indent=4)



from .constants import LBRACKET, RBRACKET, METADATA
from .data import *
from .metadata import *

from typing import Union


def make_entry(
    json_dict: dict,
) -> Union[d_entry, md_entry]:
    if METADATA in json_dict:
        # Metadata Record
        if Bracket.is_superclass(json_dict):
            return Bracket.cons(json_dict)
        else:
            assert EOF.is_superclass(json_dict)
            return EOF.cons(json_dict)
    else:
        # Data Record
        if DataVal.is_superclass(json_dict):
            return DataVal.cons(json_dict)
        elif DataFrame.is_superclass(json_dict):
            return DataFrame.cons(json_dict)
        elif DataRef.is_superclass(json_dict):
            return DataFrame.cons(json_dict)
        else:
            assert Torch.is_superclass(json_dict)
            return Torch.cons(json_dict)  # type: ignore


del Union



from . import flags
from .logger import log, pinned
from .kits import MTK
from .query import *
from .hlast import apply
from .shelf.home_shelf import get_ref
from .utils import plot_confusion_matrix
from .constants import *

flags.Parser.parse()



from .seemblock import SeemBlock

from flor import flags
from flor.journal.entry import *
from flor.state import State

import pandas as pd

from typing import List, Dict


class ReadBlock(SeemBlock):
    dynamic_identifiers: Dict[str, int] = dict()
    pda: List[Bracket] = []

    @staticmethod
    def step_into(block_name: str, probed=None):
        if probed is None:
            probed = flags.PID.ngpus > 1
        assert isinstance(block_name, str)
        dynamic_id = ReadBlock.dynamic_identifiers.get(block_name, 0)
        ReadBlock.dynamic_identifiers[block_name] = dynamic_id + 1

        lbracket = Bracket(
            block_name, dynamic_id, LBRACKET, predicate=not flags.RESUMING and probed
        )
        ReadBlock.pda.append(lbracket)
        return lbracket.predicate

    @staticmethod
    def end(*args, values=None):
        lbracket = ReadBlock.pda.pop()
        block = (
            State.target_block
            if flags.PID.pid == 0
            else ReadBlock.journal.as_tree()[lbracket.sk].blocks[lbracket.gk]
        )
        assert block is not None
        if not lbracket.predicate:
            for data_record, arg in zip(block.data_records, args):
                data_record.make_val()
                value_serialized = data_record.value
                if hasattr(arg, "load_state_dict"):
                    # PyTorch support
                    arg.load_state_dict(value_serialized)
                elif isinstance(arg, pd.DataFrame):
                    assert value_serialized is not None
                    arg.update(value_serialized)
                    arg = arg[value_serialized.columns]
                    arg = arg[: len(value_serialized)]
                elif isinstance(arg, list):
                    assert isinstance(value_serialized, list)
                    arg[:] = value_serialized
                elif isinstance(arg, dict):
                    assert isinstance(value_serialized, dict)
                    arg.update(value_serialized)
                elif hasattr(arg, "__dict__"):
                    if isinstance(value_serialized, dict):
                        arg.__dict__.update(arg)
                    else:
                        assert type(arg) == type(value_serialized)
                        arg.__dict__.update(value_serialized.__dict__)  # type: ignore
                else:
                    # TODO: ...
                    raise RuntimeError(
                        "TODO: add hooks for user-defined de-serialization"
                    )
        # TODO: ...
        assert values is None, "TODO: Add support for literals/atomics"



import json
from pathlib import PurePath
from typing import Union
import uuid

import pandas as pd

from flor.shelf import home_shelf

from ..constants import *
from .abstract import Data


class DataFrame(Data):
    def __init__(self, sk, gk, v=None, r: Union[None, PurePath] = None):
        assert bool(v is not None) != bool(r is not None)
        super().__init__(sk, gk, v)
        self.ref = r
        self.val_saved = v is None and r is not None

    def make_val(self):
        assert self.ref is not None
        self.value = pd.read_csv(self.ref)

    def would_mat(self):
        """
        For timing serialization costs
        """
        return

    def jsonify(self):
        assert self.ref is not None
        assert (
            self.val_saved and self.ref.suffix == ".csv"
        ), "Must call DataFrame.set_ref_and_dump(...) before jsonify()"
        d = super().jsonify()
        del d[VAL]
        d["csv_ref"] = str(self.ref)
        return d

    def set_ref(self, csv_ref: PurePath):
        self.ref = csv_ref

    def dump(self):
        assert self.value is not None
        assert (
            isinstance(self.ref, PurePath) and self.ref.suffix == ".csv"
        ), "Must first set a reference path with a `.csv` suffix"
        self.value.to_csv(self.ref)
        self.val_saved = True
        self.value = None

    def set_ref_and_dump(self, csv_ref: PurePath):
        self.set_ref(csv_ref)
        self.dump()

    @staticmethod
    def is_superclass(json_dict: dict):
        return "csv_ref" in json_dict or (
            "ref" in json_dict and PurePath(json_dict["ref"]).suffix == ".csv"
        )

    @classmethod
    def cons(cls, json_dict: dict):
        return cls(
            json_dict[STATIC_KEY], json_dict[GLOBAL_KEY], v=None, r=json_dict["csv_ref"]
        )

    def promise(self):
        ref = home_shelf.get_csv_ref()
        assert ref is not None
        self.set_ref_and_dump(ref)
        self.promised = self.jsonify()

    def fulfill(self):
        super().fulfill()
        return json.dumps(self.promised)



from flor.constants import *


def data_prep_pivot(df, data_prep_names):
    rolling_dataframe = None
    for dpname in data_prep_names:
        if dpname == "tstamp":
            continue  # Avoiding redundant columns
        pivot_value = df[df["name"] == dpname][
            list(DATA_PREP)
            + [
                "value",
            ]
        ]
        pivot_f = pivot_value.rename(columns={"value": dpname})
        if rolling_dataframe is None:
            rolling_dataframe = pivot_f
        else:
            rolling_dataframe = rolling_dataframe.merge(
                pivot_f, how="outer", on=list(DATA_PREP)
            )
    if rolling_dataframe is not None:
        rolling_dataframe = rolling_dataframe.drop_duplicates(subset=list(DATA_PREP))
        return rolling_dataframe


def outer_loop_pivot(df, outer_loop_names):
    rolling_dataframe = None
    for ol_name in outer_loop_names:
        if ol_name == "tstamp":
            continue  # Avoiding redundant columns
        pivot_value = df[df["name"] == ol_name][
            list(OUTR_LOOP)
            + [
                "value",
            ]
        ]
        pivot_f = pivot_value.rename(columns={"value": ol_name})
        if rolling_dataframe is None:
            rolling_dataframe = pivot_f
        else:
            rolling_dataframe = rolling_dataframe.merge(
                pivot_f, how="outer", on=list(OUTR_LOOP)
            )
    if rolling_dataframe is not None:
        rolling_dataframe = rolling_dataframe.drop_duplicates(subset=list(OUTR_LOOP))
        return rolling_dataframe


def inner_loop_pivot(df, inner_loop_names):
    rolling_dataframe = None
    for il_name in inner_loop_names:
        if il_name == "tstamp":
            continue  # Avoiding redundant columns
        pivot_value = df[df["name"] == il_name][
            list(INNR_LOOP)
            + [
                "value",
            ]
        ]
        pivot_value["step"] = pivot_value["step"].map(
            lambda x: max(int(x), 0), na_action="ignore"
        )
        pivot_f = pivot_value.rename(columns={"value": il_name})
        if rolling_dataframe is None:
            rolling_dataframe = pivot_f
        else:
            rolling_dataframe = rolling_dataframe.merge(
                pivot_f, how="outer", on=list(INNR_LOOP)
            )
    if rolling_dataframe is not None:
        rolling_dataframe = rolling_dataframe.drop_duplicates(
            subset=list(INNR_LOOP)
        )
        return rolling_dataframe



from abc import ABC, abstractmethod


class Future(ABC):
    def __init__(self, v):
        self.value = v
        self.promised = None

    @abstractmethod
    def promise(self):
        """
        Synchronous move or copy for data protection
            Example:
            self.promised = deepcopy(self.value)
        """
        ...

    @abstractmethod
    def fulfill(self) -> str:
        """
        Finishes intended work on stored version of self.value
        """
        assert self.promised is not None
        return ""



import json
from copy import deepcopy
from pathlib import PurePath
from typing import Union

import cloudpickle

from flor.shelf import home_shelf

from ..constants import *
from .abstract import Data


class Reference(Data):
    def __init__(self, sk, gk, v=None, r: Union[None, PurePath] = None):
        assert bool(v is not None) != bool(r is not None)
        super().__init__(sk, gk, v)
        self.ref = r
        self.val_saved = v is None and r is not None

    def make_val(self):
        assert self.ref is not None
        with open(self.ref, "rb") as f:
            self.value = cloudpickle.load(f)

    def would_mat(self):
        """
        For timing serialization costs
        """
        cloudpickle.dumps(self.value)

    def jsonify(self):
        assert self.ref is not None
        assert (
            self.val_saved and self.ref.suffix == PKL_SFX
        ), "Must call Reference.set_ref_and_dump(...) before jsonify()"
        d = super().jsonify()
        del d[VAL]
        d[REF] = str(self.ref)
        return d

    def set_ref(self, pkl_ref: PurePath):
        self.ref = pkl_ref

    def dump(self):
        assert (
            isinstance(self.ref, PurePath) and self.ref.suffix == PKL_SFX
        ), "Must first set a reference path with a `.pkl` suffix"
        with open(self.ref, "wb") as f:
            cloudpickle.dump(self.value, f)
        self.val_saved = True
        self.value = None

    def set_ref_and_dump(self, pkl_ref: PurePath):
        self.set_ref(pkl_ref)
        self.dump()

    @staticmethod
    def is_superclass(json_dict: dict):
        return REF in json_dict

    @classmethod
    def cons(cls, json_dict: dict):
        return cls(
            json_dict[STATIC_KEY], json_dict[GLOBAL_KEY], v=None, r=json_dict[REF]
        )

    def promise(self):
        self.promised = deepcopy(self.value)
        self.value = self.promised

    def fulfill(self):
        super().fulfill()
        ref = home_shelf.get_pkl_ref()
        assert ref is not None
        self.set_ref_and_dump(ref)
        return json.dumps(self.jsonify())



from ..constants import *
from ..abstract import Entry
from ....logger import Future

from abc import ABC, abstractmethod


class Data(Entry, Future, ABC):
    def __init__(self, sk, gk, v):
        Entry.__init__(self, sk, gk)
        Future.__init__(self, v)
        self.value = v

    def is_left(self):
        return False

    def is_right(self):
        return True

    def jsonify(self):
        d = super().jsonify()
        d[VAL] = self.value
        return d

    @abstractmethod
    def make_val(self):
        ...

    @abstractmethod
    def would_mat(self):
        ...



from flor.state import State
from flor.constants import LOG_RECORDS
from flor import flags
from typing import Optional, List, Dict, Any

from flor.query import database

import csv
import pandas as pd
from pathlib import Path, PurePath

replay_logs: List[Dict[str, Any]] = []
record_logs: List[Dict[str, Any]] = []


def deferred_init():
    with open(LOG_RECORDS, "r") as f:
        replay_logs.extend(list(csv.DictReader(f)))


def put(name, value):
    assert State.epoch is not None
    d = {
        "epoch": int(State.epoch),
        "step": int(State.step) if State.step is not None else None,
        "name": name,
        "value": value,
    }
    record_logs.append(d)


def put_dp(name, value):
    d = {
        "epoch": -1,
        "step": -1,
        "name": name,
        "value": value,
    }
    record_logs.append(d)


def get(name):
    return [d for d in replay_logs if d["name"] == name]


def flush(projid: str, tstamp: str):
    if flags.NAME and not flags.REPLAY:
        if len(record_logs) > 0:
            pd.DataFrame(record_logs).to_csv(LOG_RECORDS, index=False)
            print("Flor wrote log records locally.")
    elif flags.NAME and flags.REPLAY:
        assert State.repo is not None
        for rlg in record_logs:
            rlg["tstamp"] = str(PurePath(tstamp).stem)
            rlg["projid"] = projid
            rlg["runid"] = flags.NAME
            for version in State.repo.iter_commits():
                if "RECORD::" in str(version.message):
                    rlg["vid"] = str(version.hexsha)
                    break
            assert (
                "vid" in rlg
            ), "Failed to find a recent FLOR commit, are you calling from a flor.shadow branch?"
        database.write_log_records(record_logs)


def exists():
    return Path(LOG_RECORDS).exists()


def eval(select, where=None) -> pd.DataFrame:
    assert replay_logs is not None
    df = pd.DataFrame(replay_logs)
    if where is not None:
        df = df.query(where)
    return df[select]



from .seemblock import SeemBlock
from .readblock import ReadBlock
from .writeblock import WriteBlock

from flor import flags


class SkipBlock(SeemBlock):
    @staticmethod
    def step_into(block_name: str, probed=None):
        if flags.NAME is not None:
            raise RuntimeError("SkipBlock missing dynamic linking")
        return True

    @staticmethod
    def end(*args, values=None):
        if flags.NAME is not None:
            raise RuntimeError("SkipBlock missing dynamic linking")

    @staticmethod
    def bind():
        block = ReadBlock if flags.REPLAY else WriteBlock
        SkipBlock.step_into = block.step_into  # type: ignore
        SkipBlock.end = block.end



#!/usr/bin/env python
# type: ignore

from argparse import ArgumentParser, FileType, Namespace
from ast import AST, Name, iter_fields, parse, unparse, walk, stmt, literal_eval
from collections import defaultdict
from copy import deepcopy
from typing import Optional, TextIO
import sys
import ast

from flor.hlast.gumtree import GumTree, Mapping, python

# https://github.com/PyCQA/pylint/issues/3882
# pylint: disable=unsubscriptable-object


def add_arguments(parser: ArgumentParser):
    parser.add_argument("lineno", type=int)
    parser.add_argument("source", type=FileType("r"))
    parser.add_argument("target", type=FileType("r+"))
    parser.add_argument("--out", type=FileType("w"), default=sys.stdout)
    parser.add_argument("--minor", type=int, default=sys.version_info[1])
    parser.add_argument("--gumtree", type=literal_eval, default="{}")
    return parser


def propagate(args: Namespace):
    tree, target = [parse(f.read()) for f in (args.source, args.target)]
    args.source.close()
    args.target.close()
    args.out = open(args.out, "w")

    # llv = LogLinesVisitor()
    # llv.visit(tree)
    # fft = FlorFreeTransformer()
    # target = fft.visit(target)
    replicate(tree, find(tree, lineno=args.lineno), target, **args.gumtree)  # type: ignore
    print(unparse(target), file=args.out)


def replicate(tree: AST, node: stmt, target: AST, **kwargs):
    """`
    First we do code-block alignment using the GumTree
    algorithm from Falleri et al.
    """
    adapter = python.Adapter(tree, target)
    mapping = GumTree(adapter, **kwargs).mapping(tree, target)
    # asserting `tree` is the root of `node` in the `adapter`
    # assert tree == adapter.root(node) and isinstance(node, stmt)

    """
    Then we insert the back-propagated statement into the target block
    QUERY:
        Is the back-propagation always intra-block, meaning from
        same block in version v to same block in version ancestor(v)


    It's working. It's not off by one, it's finding an 
    adjacent injection site.

    The contextual copy ignores content of target
    """
    block, index = find_insert_loc(adapter, node, mapping)
    assert block is not None
    # if node in mapping:
    lev = LoggedExpVisitor()
    lev.visit(node)
    pnv = PairNodeVisitor(lev.name)

    if node in mapping:
        # ABORT
        edon = mapping[node]

        original = block.pop(index)  # type: ignore
        original_s = deepcopy(original)
        original = pnv.visit(node, original)
        assert pnv.success
        block.insert(index, original)  # type: ignore
    else:
        new = make_contextual_copy(adapter, node, mapping)
        block.insert(index, new)  # type: ignore

    # print("done")


class FlorFreeTransformer(ast.NodeTransformer):
    def visit_Call(self, node: ast.Call):
        pred = (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.value.id == "flor"
            and node.func.attr == "log"
        )
        if pred:
            return self.visit(node.args[1])
        return self.generic_visit(node)


class LogLinesVisitor(ast.NodeVisitor):
    def __init__(self):
        super().__init__()
        self.coreline = None
        self.linenos = set([])

    def visit_Assign(self, node: ast.Assign):
        self.coreline = node.lineno
        return self.generic_visit(node)

    def visit_AugAssign(self, node: ast.AugAssign):
        self.coreline = node.lineno
        return self.generic_visit(node)

    def visit_Expr(self, node: ast.Expr):
        self.coreline = node.lineno
        return self.generic_visit(node)

    def visit_Call(self, node: ast.Call):
        pred = (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.value.id == "flor"
            and node.func.attr == "log"
        )
        if pred:
            self.linenos.add(self.coreline)

        return self.generic_visit(node)


class PairNodeVisitor(ast.NodeTransformer):
    def __init__(self, name):
        super().__init__()
        self.name = name
        self.success = False

    def make_wrapper(self, child):
        if isinstance(child, AST):
            return (
                ast.parse(f"flor.log('{self.name}',{unparse(child).strip()})")
                .body[0]
                .value  # type: ignore
            )
        elif isinstance(child, (list, tuple)):
            return (
                ast.parse(
                    f"flor.log('{self.name}',{[unparse(c).strip() for c in child]})"
                )
                .body[0]
                .value  # type:ignore
            )
        else:
            raise NotImplementedError

    def equals(self, node1: AST, node2: AST):
        if type(node1) != type(node2):
            return False
        for a in node1._fields:
            if type(getattr(node1, a)) != type(getattr(node2, a)):
                return False
        return True

    def visit(self, node1: AST, node2: AST):
        """Visit a node."""
        if not self.equals(node1, node2):

            field = None

            for f in node1._fields:
                v = getattr(node1, f)
                s = unparse(v).strip() if isinstance(v, AST) else str(v)
                if "flor.log" in s:
                    field = f
                    break
            assert field is not None
            child = getattr(node2, field)
            logging_child = self.make_wrapper(child)
            setattr(node2, field, logging_child)
            self.success = True
            return node2
        return self.generic_visit(node1, node2)

    def generic_visit(self, node1, node2):
        for (fld1, old_val1), (fld2, old_val2) in zip(
            iter_fields(node1), iter_fields(node2)
        ):
            if isinstance(old_val1, list):
                assert isinstance(old_val2, list)
                new_values = []
                for val1, val2 in zip(old_val1, old_val2):
                    if isinstance(val1, AST):
                        assert isinstance(val2, AST)
                        val2 = self.visit(val1, val2)
                        if val2 is None:
                            continue
                        elif not isinstance(val2, AST):
                            new_values.extend(val2)
                            continue
                    new_values.append(val2)
                old_val2[:] = new_values
            elif isinstance(old_val1, AST):
                assert isinstance(old_val2, AST)
                new_node = self.visit(old_val1, old_val2)
                if new_node is None:
                    delattr(node1, fld1)
                    delattr(node2, fld2)
                else:
                    setattr(node2, fld2, new_node)
        return node2


class LoggedExpVisitor(ast.NodeVisitor):
    def __init__(self):
        super().__init__()
        self.name: Optional[str] = None

    def visit_Call(self, node: ast.Call):
        pred = (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.value.id == "flor"
            and node.func.attr == "log"
        )
        if not pred:
            return self.generic_visit(node)
        if len(node.args) == 2 and isinstance(node.args[0], ast.Constant):
            self.name = str(node.args[0].value)
        else:
            raise IndexError("FLOR: Did you give flor.log a key? It takes 2 args.")


def find_insert_loc(adapter, node, mapping):
    parent = adapter.parent(node)

    context = None
    for sibling in adapter.children(parent):
        if id(sibling) == id(node):
            break
        if sibling in mapping:
            context = sibling

    if context is not None:
        ref = mapping[context]
        block = adapter.parent(ref)
        index = 1 + block.index(ref)
    elif parent in mapping:
        block = mapping[parent]
        index = 0
    else:
        exit("Unable to map context!")

    return block, index


def make_contextual_copy(adapter, node, mapping):
    renames = defaultdict(lambda: defaultdict(int))
    for source, target in mapping.items():
        if isinstance(source, Name) and not adapter.contains(source, node):
            renames[source.id][target.id] += 1

    new = deepcopy(node)
    for n in walk(new):
        if isinstance(n, Name) and n.id in renames:
            n.id = max(renames[n.id], key=renames[n.id].get)  # type: ignore
    return new


def find(t: AST, *, lineno: int):
    adapter = python.Adapter(t)  # FIXME: odd dependency
    res = None
    for n in adapter.postorder(t):
        if getattr(n, "lineno", lineno) == lineno and isinstance(n, stmt):
            res = n
    return res


if __name__ == "__main__":
    propagate(add_arguments(ArgumentParser()).parse_args(sys.argv[1:]))



from typing import Dict
import ast


class LoggedExpVisitor(ast.NodeVisitor):
    def __init__(self):
        super().__init__()
        self.names: Dict[str, int] = {}

    def visit_Call(self, node: ast.Call):
        pred = (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.value.id == "flor"
            and node.func.attr == "log"
        )
        if not pred:
            return self.generic_visit(node)
        if len(node.args) == 2 and isinstance(node.args[0], ast.Constant):
            self.names[str(node.args[0].value)] = node.lineno
        else:
            raise IndexError("FLOR: Did you give flor.log a key? It takes 2 args.")


class NoGradVisitor(ast.NodeVisitor):
    def __init__(self):
        super().__init__()
        self.feeding = False
        self.names = {}
        self.tree = None

    def visit_Call(self, node: ast.Call):
        pred = (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.value.id == "flor"
            and node.func.attr == "log"
        )
        if not pred or not self.feeding:
            return self.generic_visit(node)
        if len(node.args) == 2 and isinstance(node.args[0], ast.Constant):
            self.names[str(node.args[0].value)] = node.lineno
        else:
            raise IndexError("FLOR: Did you give flor.log a key? It takes 2 args.")

    def visit_With(self, node: ast.With):
        if [True for each in node.items if "torch.no_grad" in ast.unparse(each)]:
            try:
                feeding = self.feeding
                self.feeding = True
                self.tree = node
                for stmt in node.body:
                    self.visit(stmt)
            finally:
                self.feeding = feeding  # type: ignore

class NoGradTransformer(ast.NodeTransformer):
    def __init__(self, their_tree) -> None:
        super().__init__()
        self.their_tree = their_tree

    def visit_With(self, node: ast.With):
        if [True for each in node.items if "torch.no_grad" in ast.unparse(each)]:
            return self.generic_visit(self.their_tree)
        else:
            return self.generic_visit(node)
    


from ..constants import *
from .abstract import Metadata

VALID_BRACKETS = (LBRACKET, RBRACKET)


class Bracket(Metadata):
    def __init__(self, sk, gk, meta, predicate=None, timestamp=None):
        assert meta in VALID_BRACKETS
        super().__init__(sk, gk, meta)
        self.predicate = predicate
        self.timestamp = timestamp

    def is_left(self):
        return self.meta == LBRACKET

    def is_right(self):
        return self.meta == RBRACKET

    @staticmethod
    def is_superclass(json_dict: dict):
        return METADATA in json_dict and json_dict[METADATA] in VALID_BRACKETS

    @classmethod
    def cons(cls, json_dict: dict):
        return cls(json_dict[STATIC_KEY], json_dict[GLOBAL_KEY], json_dict[METADATA])



from ..constants import *
from ..abstract import Entry
from ....logger import Future

from abc import ABC
import json


class Metadata(Entry, Future, ABC):
    def __init__(self, sk, gk, meta):
        Entry.__init__(self, sk, gk)
        self.meta = meta

    def jsonify(self):
        d = super().jsonify()
        d[METADATA] = self.meta
        return d

    def promise(self):
        ...

    def fulfill(self):
        return json.dumps(self.jsonify())



from pathlib import PurePath, Path
from typing import Any, Dict
from flor.shelf import cwd_shelf
from flor.state import State
from flor.constants import *
from flor import flags

from flor.query import database

import os
import shutil
import json
import csv
import pandas as pd


def clear_stash():
    global stash
    (Path.home() / ".flor").mkdir(exist_ok=True)
    (Path.home() / ".flor" / cwd_shelf.get_projid()).mkdir(exist_ok=True)
    stash = Path.home() / ".flor" / cwd_shelf.get_projid() / "stash"
    if stash.exists():
        shutil.rmtree(stash)
    stash.mkdir()
    return stash


def filtered_versions():
    r = State.repo
    assert r is not None

    all_versions = [version for version in r.iter_commits()]
    flor_versions = [
        version for version in all_versions if "::" in str(version.message)
    ]
    record_versions = [
        version for version in flor_versions if "RECORD::" in str(version.message)
    ]

    return {"ALL": all_versions, "FLOR": flor_versions, "RECORD": record_versions}


def resolve_cache(cache_short_path):
    assert cwd_shelf.in_shadow_branch()
    return Path.home() / ".flor" / cwd_shelf.get_projid() / cache_short_path


def unpack():
    """
    MAIN FUNCTION
    """
    assert cwd_shelf.in_shadow_branch()
    clear_stash()

    r = State.repo
    assert r is not None
    if State.db_conn is None:
        database.start_db(cwd_shelf.get_projid())
    wmrk = database.get_watermark()
    active_branch = State.active_branch
    try:
        commits = filtered_versions()
        for version in commits["RECORD"]:
            if wmrk is not None and version.hexsha in wmrk:
                break
            try:
                # print(f"STEPPING IN {version.hexsha}")
                r.git.checkout(version)
                cp_seconds(version)
                cp_log_records(version)
            except Exception as e:
                print(e)
    finally:
        r.git.checkout(active_branch)


def cp_seconds(version):
    assert State.db_conn is not None
    hexsha, message = version.hexsha, version.message
    if "RECORD::" in message and SECONDS_JSON.exists():
        with open(REPLAY_JSON, "r") as f:
            replay_json = json.load(f)
        with open(SECONDS_JSON, "r") as f:
            seconds_json = json.load(f)
        assert isinstance(seconds_json, dict) and seconds_json
        assert "PREP" in seconds_json

        def prep_normalize(prep_secs):
            nonlocal seconds_json, replay_json
            data = []
            data.append(
                {
                    c: v
                    for c, v in zip(
                        list(DATA_PREP)
                        + [
                            "seconds",
                        ],
                        [
                            cwd_shelf.get_projid(),
                            replay_json["NAME"],
                            PurePath(replay_json["TSTAMP"]).stem,
                            hexsha,
                            float(prep_secs),
                        ],
                    )
                }
            )
            return data

        def outr_normalize(all_epochs_secs):
            nonlocal seconds_json, replay_json
            data = []
            for i, epoch_secs in enumerate(all_epochs_secs):
                epoch = i + 1
                data.append(
                    {
                        c: v
                        for c, v in zip(
                            list(OUTR_LOOP)
                            + [
                                "seconds",
                            ],
                            [
                                cwd_shelf.get_projid(),
                                replay_json["NAME"],
                                PurePath(replay_json["TSTAMP"]).stem,
                                hexsha,
                                int(epoch),
                                float(epoch_secs),
                            ],
                        )
                    }
                )
            return data

        # ..send PREP to data_prep
        pd.DataFrame(prep_normalize(seconds_json["PREP"])).to_sql(
            "data_prep", con=State.db_conn, if_exists="append", index=False
        )

        # .. send EPOCHS to outr_loop
        if "EPOCHS" in seconds_json:
            pd.DataFrame(outr_normalize(seconds_json["EPOCHS"])).to_sql(
                "outr_loop", con=State.db_conn, if_exists="append", index=False
            )


def cp_log_records(version):
    assert State.db_conn is not None
    hexsha, message = version.hexsha, version.message
    if PurePath(message).suffix == ".json":
        # Older  Versions
        messages = message.split("::")
        tstamp_json = messages[-1]
        tstamp_json = PurePath(tstamp_json)
        replay_json = get_replay_json()
        if replay_json is not None:
            lr_csv = get_log_records_csv()
            data = normalize(replay_json, lr_csv, hexsha, tstamp_json)
            df = pd.DataFrame(data)
            df.to_csv(stash / tstamp_json.with_suffix(".csv"), index=False)
            df.to_sql("log_records", con=State.db_conn, if_exists="append", index=False)
    else:
        # Newer Versions, Non-Flor Versions
        replay_json = get_replay_json()
        if replay_json is not None:
            lr_csv = get_log_records_csv()
            tstamp_json = get_tstamp_json(replay_json)
            if tstamp_json is not None:
                data = normalize(replay_json, lr_csv, hexsha, tstamp_json)
                df = pd.DataFrame(data)
                df.to_sql(
                    "log_records", con=State.db_conn, if_exists="append", index=False
                )


def get_replay_json():
    candidate_paths = {">=2.5.1": REPLAY_JSON, "<2.5": Path(".replay.json")}
    for vs, p in candidate_paths.items():
        if p.exists():
            return p
    return None


def get_log_records_csv():
    candidate_paths = {">=2.5.1": LOG_RECORDS, "<2.5": Path("log_records.csv")}
    for vs, p in candidate_paths.items():
        if p.exists():
            return p
    return None


def normalize(replay_json, lr_csv, hexsha, tstamp):
    """
    Draw data from replay_json.kvs, replay_json.data_prep, lr_csv
    """
    assert replay_json is not None
    data = []

    with open(replay_json, "r") as f:
        d: Dict[str, Any] = json.load(f)

    user_vars = [name for name in d if not name.isupper()]
    for user_var in user_vars:
        # append p,v,-1,-1,n,v
        data.append(
            {
                "projid": cwd_shelf.get_projid(),
                "runid": d["NAME"],
                "tstamp": tstamp.stem,
                "vid": hexsha,
                "epoch": -1,
                "step": -1,
                "name": user_var,
                "value": d[user_var],
            }
        )
    if "KVS" in d:
        # LEGACY
        _kvs = d["KVS"]

        for k in _kvs:
            if len(k.split(".")) >= 3:
                z = k.split(".")
                e = z.pop(0)
                _ = z.pop(0)
                n = ".".join(z)
                for s, x in enumerate(_kvs[k]):
                    # pvresnx
                    # seq.append((d["NAME"], d["MEMO"], tstamp, r, e, s, n, x))
                    data.append(
                        {
                            "projid": cwd_shelf.get_projid(),
                            "runid": d["NAME"],
                            "tstamp": tstamp.stem,
                            "vid": hexsha,
                            "epoch": int(e) if e else -1,
                            "step": int(s) if s else -1,
                            "name": n,
                            "value": x,
                        }
                    )
    if lr_csv is not None:
        with open(lr_csv, "r") as f:
            data.extend(
                [
                    {
                        "projid": cwd_shelf.get_projid(),
                        "runid": d["NAME"],
                        "tstamp": tstamp.stem,
                        "vid": hexsha,
                        "epoch": int(each["epoch"]) if each["epoch"] else -1,
                        "step": int(each["step"]) if each["step"] else -1,
                        "name": str(each["name"]),
                        "value": each["value"],
                    }
                    for each in csv.DictReader(f)
                ]
            )
    return data


def get_tstamp_json(replay_json):
    assert replay_json is not None
    with open(replay_json, "r") as f:
        kvs = json.load(f)
    if "TSTAMP" in kvs:
        return PurePath(kvs["TSTAMP"]).with_suffix(".json")
    elif "tstamp" in kvs:
        return PurePath(kvs["tstamp"]).with_suffix(".json")



from ..constants import *
from .abstract import Metadata

from typing import List


class EOF(Metadata):
    def __init__(self, sparse: List[int], itc: int, commit_sha: str):
        super().__init__(None, None, EOF_NAME)
        self.sparse_checkpoints = sparse
        self.iterations_count = itc
        self.commit_sha = commit_sha

    def is_left(self):
        return False

    def is_right(self):
        return False

    def jsonify(self):
        d = dict()
        d[METADATA] = EOF_NAME
        d[SPARSE_CHECKPOINTS] = self.sparse_checkpoints
        d[ITERATIONS_COUNT] = int(self.iterations_count)
        d["COMMIT_SHA"] = self.commit_sha
        return d

    @staticmethod
    def is_superclass(json_dict: dict):
        return METADATA in json_dict and json_dict[METADATA] == EOF_NAME

    @classmethod
    def cons(cls, json_dict: dict):
        sparse = json_dict[SPARSE_CHECKPOINTS]
        assert isinstance(sparse, List)
        return cls(sparse, int(json_dict[ITERATIONS_COUNT]), json_dict["COMMIT_SHA"])



from ..journal import Journal
from ..logger import Logger

from abc import ABC, abstractmethod


class SeemBlock(ABC):
    journal = Journal()
    logger = Logger()

    @staticmethod
    @abstractmethod
    def step_into(block_name: str, probed=None):
        ...

    @staticmethod
    @abstractmethod
    def end(*args, values=None):
        ...



from .block import Block
from typing import List


class BlockGroup:
    def __init__(self, first: Block):
        self.blocks: List[Block] = [
            first,
        ]

        self.materialization_time = None
        self.computation_time = None
        self.executions_count = 0
        self.materializations_count = 0

    def tick_execution(self, t):
        self.executions_count += 1
        if self.computation_time is None:
            self.computation_time = t

    def should_time_mat(self):
        assert self.executions_count > 0
        return self.executions_count == 1

    def tick_materialization(self):
        self.materializations_count += 1

    def set_mat_time(self, t):
        if self.materialization_time is None:
            self.materialization_time = t

    def add_block(self, block: Block):
        self.blocks.append(block)

    def peek_block(self) -> Block:
        return self.blocks[-1]



import glob
import os
from pathlib import Path
from typing import List, Generator

from .. import shelf
from .future import Future
from .. import flags


class Logger:
    def __init__(self, path=None, buf_size=None):
        self.path = Path(path) if path is not None else None
        self.buffer = Buffer() if buf_size is None else Buffer(buf_size)
        self.flush_count = 0

    def set_path(self, path):
        self.path = Path(path)

    def append(self, o: Future):
        assert self.path is not None, "Logger path not set."
        self.buffer.append(o)
        if self.buffer.is_full():
            self.flush()

    def flush(self, is_final=False):
        assert self.path is not None
        self.flush_count += 1
        # pid = os.fork()
        # if not pid:
        self._flush_buffer()
        if is_final:
            p = self.path.with_name(self.path.stem + "_*" + self.path.suffix)
            with open(self.path, "wb") as out_f:
                for pi in glob.glob(p.as_posix()):
                    with open(pi, "rb") as in_f:
                        out_f.write(in_f.read())
                    os.remove(pi)
        # else:
        self.buffer.clear()

    def force(self):
        self._flush_buffer()

    def _flush_buffer(self):
        assert isinstance(self.path, Path)
        p = self.path.with_name(
            self.path.stem + f"_{self.flush_count}" + self.path.suffix
        )
        with open(p, "w") as f:
            for o in self.buffer.flush():
                assert o is not None
                f.write(o + os.linesep)


class Buffer:
    def __init__(self, size=1024):
        self._b: List[Future] = []
        self.size = size

    def append(self, o: Future):
        assert isinstance(
            o, Future
        ), "object passed to flor Logger must implement Future interface"
        o.promise()
        self._b.append(o)

    def is_full(self):
        return len(self._b) >= self.size

    def clear(self):
        self._b[:] = []

    def flush(self) -> Generator[str, None, None]:
        for o in self._b:
            yield o.fulfill()
        self.clear()

    def __len__(self):
        return len(self._b)



from flor.journal.entry import *
from typing import Union, List, get_args


class Block:
    def __init__(self, log_record: md_entry, parent=None):
        assert log_record.is_left()

        self.child: Union["Block", None] = None
        self.parent: Union["Block", None] = parent
        self.successor: Union["Block", None] = None

        self.static_key = log_record.sk
        self.global_key = log_record.gk
        self.data_records: List[d_entry] = []
        self.right_fed = False
        self.force_mat = False

    def belongs_in_block(self, data_record: Union[d_entry, md_entry]):
        assert data_record.is_right()
        return data_record.sk == self.static_key and data_record.gk == self.global_key

    def feed_entry(self, data_record: Union[d_entry, md_entry]):
        assert self.belongs_in_block(data_record)
        self.right_fed = True
        if not isinstance(data_record, get_args(md_entry)):
            self.data_records.append(data_record)  # type: ignore

    def force_mat_successors(self):
        assert self.parent is None
        first_static_key = self.static_key
        block = self.successor
        while block is not None and block.static_key != first_static_key:
            block.force_mat = True
            block = block.successor



import ast
from argparse import Namespace
from os import PathLike
from shutil import copyfile
from pathlib import Path, PurePath
from sys import stdout
from typing import Dict, List, Set

from flor.hlast.gtpropagate import propagate, LogLinesVisitor  # type: ignore
from flor.state import State
import flor.query as q
from .visitors import LoggedExpVisitor, NoGradVisitor, NoGradTransformer


def backprop(lineno: int, source: str, target: str, out=None):
    syntactic_prop(lineno, source, target, out)


def syntactic_prop(lineno: int, source, target, out=None):
    if out is None:
        with open(str(source), "r") as src, open(str(target), "r") as dst:
            return propagate(
                Namespace(
                    lineno=lineno,
                    source=src,
                    target=dst,
                    out=str(target),
                    gumtree=dict(),
                )
            )
    else:
        with open(str(source), "r") as src, open(str(target), "r") as dst:
            return propagate(
                Namespace(
                    lineno=lineno, source=src, target=dst, out=out, gumtree=dict()
                )
            )


class StmtToPropVisitor(ast.NodeVisitor):
    def __init__(self, lineno) -> None:
        super().__init__()
        self.value = ""
        self.value_valid = False
        self.lineno = int(lineno)

    def generic_visit(self, node: ast.AST):
        if isinstance(node, ast.stmt):
            assert node.end_lineno is not None
            if int(node.lineno) == int(self.lineno):
                self.value = str(ast.unparse(node))
                self.value_valid = True
            else:
                super().generic_visit(node)
        else:
            super().generic_visit(node)


def apply(names: List[str], dst: str):
    fp = Path(dst)
    facts = q.log_records(skip_unpack=True)
    # Get latest timestamp for each variable name
    valid_names = facts[facts["name"].isin(names)][["name", "tstamp", "vid", "value"]]
    valid_names = valid_names[valid_names["value"].notna()]
    name2tstamp = valid_names[["name", "tstamp", "vid"]].drop_duplicates()

    stash = q.clear_stash()
    assert stash is not None
    assert State.repo is not None
    copyfile(fp, stash / fp)
    hits: Set[str] = set([])
    grouped_names: Dict[str, int] = {}

    for _, row in name2tstamp.iterrows():
        n = row["name"]
        v = row["vid"]
        State.repo.git.checkout(v, "--", fp)
        lev = LoggedExpVisitor()
        with open(fp, "r") as f:
            lev.visit(ast.parse(f.read()))
        if n in lev.names:
            grouped_names[n] = lev.names[n]
            hits.add(n)
            copyfile(src=fp, dst=stash / PurePath(n).with_suffix(".py"))
        if len(hits) == len(names):
            break

    copyfile(stash / fp, fp)
    assert len(hits) == len(
        names
    ), f"Failed to find log statement for vars {[n for n in names if n not in hits]}"

    # Next, from the stash you will apply each file to our main one
    parse_noGrad = []
    for name in names:
        with open(stash / PurePath(name).with_suffix(".py"), "r") as f:
            tree = ast.parse(f.read())

        ng_visitor = NoGradVisitor()
        ng_visitor.visit(tree)
        if name not in ng_visitor.names:
            lineno = int(grouped_names[name])
            # lev possibly unbound
            backprop(lineno, str(stash / PurePath(name).with_suffix(".py")), dst)
            print(f"Applied {name} to {dst}")
        else:
            parse_noGrad.append(ng_visitor.tree)

    if len(parse_noGrad) > 1:
        raise NotImplementedError(
            "TODO: Will need to merge code blocks, union of logging statements"
        )
    elif len(parse_noGrad) == 1:
        their_tree = parse_noGrad[0]
        with open(dst, "r") as f:
            my_tree = ast.parse(f.read())
        ng_transformer = NoGradTransformer(their_tree)

        with open(dst, "w") as f:
            f.write(ast.unparse(ng_transformer.visit(my_tree)))


__all__ = ["backprop", "apply"]

if __name__ == "__main__":
    backprop(78, "cases/train_rnn/now.py", "cases/train_rnn/before.py")
